-- The top module of Deep Learning on FPGA
-- by Toshio Iwata at digitalfilter.com 2019/02/13

library ieee ;
use ieee.std_logic_1164.all ;
use ieee.std_logic_arith.all ;
use ieee.std_logic_signed.all;

-------------------------------------------------------------------------------
-- Entity
-------------------------------------------------------------------------------
entity deep_hw is
port(
  CLK : in std_logic;                            -- clock (100MHz)
  RST_N : in std_logic;                          -- reset (active low)
  ANSWER : Out std_logic_vector(3 downto 0) );  
end deep_hw;

-------------------------------------------------------------------------------
-- Architecture
-------------------------------------------------------------------------------
architecture RTL of deep_hw is

component ROM_X
   Port (
          CLK : In std_logic;
          CS_N : In std_logic;
          ADDR : In std_logic_vector(12 downto 0);
          DATA : Out std_logic_vector(7 downto 0) );
end component;

component layer1
Port (
  CLK : in std_logic;                            -- clock (100MHz)
  RST_N : in std_logic;                          -- reset (active low)
  X_RD_DATA : in std_logic_vector(7 downto 0); 
  X_RD_ADDR  : Out std_logic_vector(9 downto 0);  
  Z_WR_DATA : Out std_logic_vector(7 downto 0);  
  Z_WR_ADDR  : Out std_logic_vector(5 downto 0);  
  Z_WR_N  : Out std_logic;  
  W1START : in std_logic; 
  W1DONE : Out std_logic );  
end component;

component layer2
Port (
  CLK : in std_logic;                            -- clock (100MHz)
  RST_N : in std_logic;                          -- reset (active low)
  X_RD_DATA : in std_logic_vector(7 downto 0); 
  X_RD_ADDR  : Out std_logic_vector(5 downto 0);  
  Z_WR_DATA : Out std_logic_vector(7 downto 0);  
  Z_WR_ADDR  : Out std_logic_vector(6 downto 0);  
  Z_WR_N  : Out std_logic;  
  W2START : in std_logic;
  W2DONE : Out std_logic );  
end component;

component layer3
Port (
  CLK : in std_logic;                            -- clock (100MHz)
  RST_N : in std_logic;                          -- reset (active low)
  X_RD_DATA : in std_logic_vector(7 downto 0); 
  X_RD_ADDR  : Out std_logic_vector(6 downto 0);  
  AOUT_TO_JUDGE : Out std_logic_vector(13 downto 0);  
  NUMBER_TO_JUDGE : Out std_logic_vector(3 downto 0);  
  GATE_TO_JUDGE : Out std_logic;  
  W3DONE : Out std_logic );  
end component;

component RAM_Z1
   Port (
          CLK : In std_logic;
          CS_N : In std_logic;
          WR_N : In std_logic;
          WRADDR : In std_logic_vector(5 downto 0);
          RDADDR : In std_logic_vector(5 downto 0);
          WRDATA : In std_logic_vector(7 downto 0);
          RDDATA : Out std_logic_vector(7 downto 0) );
end component;

component RAM_Z2
   Port (
          CLK : In std_logic;
          CS_N : In std_logic;
          WR_N : In std_logic;
          WRADDR : In std_logic_vector(6 downto 0);
          RDADDR : In std_logic_vector(6 downto 0);
          WRDATA : In std_logic_vector(7 downto 0);
          RDDATA : Out std_logic_vector(7 downto 0) );
end component;

component judge
  Port (
  CLK : in std_logic;                            -- clock (100MHz)
  RST_N : in std_logic;                          -- reset (active low)
  AIN : in std_logic_vector(13 downto 0); 
  NUMBER : in std_logic_vector(3 downto 0); 
  GATE : in std_logic; 
  JSTART : in std_logic; 
  ANSWER : Out std_logic_vector(3 downto 0) );  
end component;

-------------------------------------------------------------------------------
-- Signal
-------------------------------------------------------------------------------
signal X_RD_DATA : std_logic_vector(7 downto 0);
signal X_RD_ADDR : std_logic_vector(12 downto 0);

signal Z1_WR_DATA, Z1_RD_DATA, Z1_RD_DATA_A, Z1_RD_DATA_B : std_logic_vector(7 downto 0);
signal Z1_WR_ADDR, Z1_RD_ADDR : std_logic_vector(5 downto 0);
signal Z1_WR_N, Z1_WR_N_A, Z1_WR_N_B : std_logic;

signal Z2_WR_DATA, Z2_RD_DATA, Z2_RD_DATA_A, Z2_RD_DATA_B : std_logic_vector(7 downto 0);
signal Z2_WR_ADDR, Z2_RD_ADDR : std_logic_vector(6 downto 0);
signal Z2_WR_N, Z2_WR_N_A, Z2_WR_N_B : std_logic;

signal AIN : std_logic_vector(13 downto 0);
signal GATE : std_logic;
signal NUMBER  : std_logic_vector(3 downto 0);
signal PictureN  : std_logic_vector(2 downto 0);
signal W3DONE : std_logic;
signal ODD : std_logic;

begin

---------------------------------------------------------
ROM_X_inst : ROM_X port map (
  CLK => CLK,  
  CS_N => '0', 
  ADDR => X_RD_ADDR, 
  DATA => X_RD_DATA );

  X_RD_ADDR(12 downto 10) <= PictureN;

---------------------------------------------------------
layer1_inst : layer1 port map (
          CLK => CLK,
          RST_N => RST_N,
          X_RD_DATA => X_RD_DATA,
          X_RD_ADDR => X_RD_ADDR(9 downto 0),
          Z_WR_DATA => Z1_WR_DATA,
          Z_WR_ADDR => Z1_WR_ADDR,
          Z_WR_N => Z1_WR_N,
          W1START => W3DONE,
          W1DONE => open);

---------------------------------------------------------
  Z1_RD_DATA <= Z1_RD_DATA_B when ODD = '0' else Z1_RD_DATA_A;

layer2_inst : layer2 port map (
          CLK => CLK,
          RST_N => RST_N,
          X_RD_DATA => Z1_RD_DATA,
          X_RD_ADDR => Z1_RD_ADDR,
          Z_WR_DATA => Z2_WR_DATA,
          Z_WR_ADDR => Z2_WR_ADDR,
          Z_WR_N => Z2_WR_N,
          W2START => W3DONE,
          W2DONE => open);

---------------------------------------------------------
  Z2_RD_DATA <= Z2_RD_DATA_B when ODD = '0' else Z2_RD_DATA_A;

layer3_inst : layer3 port map (
          CLK => CLK,
          RST_N => RST_N,
          X_RD_DATA => Z2_RD_DATA,
          X_RD_ADDR => Z2_RD_ADDR,
          AOUT_TO_JUDGE => AIN,
          NUMBER_TO_JUDGE => NUMBER,
          GATE_TO_JUDGE => GATE,
          W3DONE => W3DONE);

---------------------------------------------------------
process( RST_N, CLK )  
begin 
	if( RST_N = '0' ) then
		PictureN <= (others => '0');
	elsif( CLK'event and CLK = '1' ) then
		if(W3DONE = '1') then 
			PictureN <= PictureN + 1;
		end if;
	end if;  
end process;

---------------------------------------------------------
process( RST_N, CLK )  
begin 
	if( RST_N = '0' ) then
		ODD <= '0';
	elsif( CLK'event and CLK = '1' ) then
		if(W3DONE = '1') then 
			ODD <= not ODD;
		end if;
	end if;  
end process;

  Z1_WR_N_A <= Z1_WR_N when ODD = '0' else '1';
  Z1_WR_N_B <= Z1_WR_N when ODD = '1' else '1';
  Z2_WR_N_A <= Z2_WR_N when ODD = '0' else '1';
  Z2_WR_N_B <= Z2_WR_N when ODD = '1' else '1';

---------------------------------------------------------
RAM_Z1_inst_A : RAM_Z1 port map (
          CLK => CLK,
          CS_N => '0',
          WR_N => Z1_WR_N_A,
          WRADDR => Z1_WR_ADDR,
          RDADDR => Z1_RD_ADDR,
          WRDATA => Z1_WR_DATA,
          RDDATA => Z1_RD_DATA_A );

RAM_Z1_inst_B : RAM_Z1 port map (
          CLK => CLK,
          CS_N => '0',
          WR_N => Z1_WR_N_B,
          WRADDR => Z1_WR_ADDR,
          RDADDR => Z1_RD_ADDR,
          WRDATA => Z1_WR_DATA,
          RDDATA => Z1_RD_DATA_B );

---------------------------------------------------------
RAM_Z2_inst_A : RAM_Z2 port map (
          CLK => CLK,
          CS_N => '0',
          WR_N => Z2_WR_N_A,
          WRADDR => Z2_WR_ADDR,
          RDADDR => Z2_RD_ADDR,
          WRDATA => Z2_WR_DATA,
          RDDATA => Z2_RD_DATA_A );

RAM_Z2_inst_B : RAM_Z2 port map (
          CLK => CLK,
          CS_N => '0',
          WR_N => Z2_WR_N_B,
          WRADDR => Z2_WR_ADDR,
          RDADDR => Z2_RD_ADDR,
          WRDATA => Z2_WR_DATA,
          RDDATA => Z2_RD_DATA_B );

---------------------------------------------------------
judge_inst : judge port map (
          CLK => CLK,
          RST_N => RST_N,
          AIN => AIN,
          NUMBER => NUMBER,
          GATE => GATE,
          JSTART => W3DONE,
          ANSWER => ANSWER );

end RTL;

